#!/usr/bin/env bash

pipe_dir=$1
pipe="$pipe_dir/nsx_container_pipe"

NSX_CMD_TIMEOUT="nsx_timeout"
NSX_CMD_OUTPUT="nsx_output"

log_level() {
    if [ ! -z "$2" ]; then
        echo [level="$1"] $2 | logger -t NSX
    fi
}

log_info() {
    log_level "INFO" "$1"
}

log_debug() {
    log_level "VERBOSE" "$1"
}

log_error() {
    log_level "ERROR" "$1"
}

bringup_vif() {
    local vif="$1"
    vif_ip=$(ifconfig $vif | grep -o -E "inet \S+" | cut -d " " -f 2 | cut -d ":" -f 2)

    if [ -z $vif_ip ]; then
        log_debug "Bringing up vif - $vif"
        timeout 20s ifdown $vif &> /dev/null

        # Kill any lingering dhclient process
        dhclient_pids=$(ps aux | grep -e "dhclient.$vif" -e "dhclient--$vif" \
                        -e "dhclient -nw $vif"| grep -v grep |\
                        tr -s ' ' ' ' | cut -d ' ' -f 2)
        for pid in $dhclient_pids; do
            log_info "Terminating dhclient process($pid) for $vif"
            kill -9 $pid
        done

        # Don't run ifup in background, it can block ifdown. Hence, timeout.
        timeout 20s ifup $vif &> /dev/null
    fi
}

close_proc_fds() {
    for fd in $(ls /proc/$$/fd); do
      case "$fd" in
        0|1|2|255)
        ;;
      *)
        eval "exec $fd>&-"
        ;;
      esac
    done
}

exec_cmd_and_timeout() {
     # command format examples
     # 1. "ifup eth0"
     # 2. "chroot .... ip netns nsx dns_watch ..."
     # 3. "nsx_timeout=120s  nsx_output=1234 chroot ... ip netns nsx support_save.py ..."
    cmd_list=$1
    timeout=""
    output_file=""

    # close any inheriented file descriptors.
    close_proc_fds
    while :
    do
        IFS='#' read cmd1 cmd2 <<< $cmd_list
        if [ -z "$cmd1" ]; then
            break;
        else
            cmd_line=$(eval echo $cmd1)
            cmd_list=$cmd2
        fi
        while :
        do
            IFS=' ' read line1 line2 <<< $cmd_line
            IFS='=' read key val <<< $line1
            if [[ $key == $NSX_CMD_TIMEOUT ]] && [[ -z $timeout ]]; then
                timeout=$val"s"
                cmd_line=$line2
            elif [[ $key == $NSX_CMD_OUTPUT ]] && [[ -z $output_file ]]; then
                output_file=$pipe_dir/$val
                cmd_line=$line2
            else
                break
            fi
        done
        log_info "parse cmd=$cmd_line, timeout=$timeout, output_file=$output_file"
        if [[ ! -z $timeout ]] && [[ -p $output_file ]] ; then
            cmd_line="timeout $timeout $cmd_line &>$output_file"
        fi
        log_info "execute cmd=$cmd_line"
        eval $cmd_line
        status=$?
        log_info "cmd=$cmd_line executed with status=$status, output_file=$output_file"
        if [ $status -ne 0 ]; then
            break
        fi
    done
    if [[ ! -z $timeout ]] && [[ -p $output_file ]] ; then
        # mark delimiter so that agent knows cmd is done
        echo status=$status >> $output_file
        printf "\x0\x0\x0\x0" >> $output_file
    fi
}

log_info "Starting NSX host agent server at $pipe"
while true
do
    if [[ -p $pipe ]]  && read line <$pipe; then
        log_info "received cmd=$line"
        exec_cmd_and_timeout "$line" & # execute the command in back ground
        disown -a # same as nohup
    else
        # read from pipe failed.
        # nsx-agent may be restarted.
        sleep 1
        log_info "waiting ...."
    fi
done
