# **********************************************************************
# Copyright 2019 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************

# This module has *nix specific code. Do not use with Windows. Module
# users should ensure compatibility with the used platform. Also if
# possible, dedup this with the same module in management_api.

from xml.dom import minidom
import xml.etree.ElementTree as ET
import os
import errno

from cli.utils.location_utils import FilePaths
from cli.utils.error_code import ErrorCode as ec
from cli.utils.node_utils import get_node_type
from cli.utils.node_utils import is_windows
from cli.utils.constants import NODE_TYPE_HOST_ESX

import nsx_logging
lg = nsx_logging.getLogger("cli.commands.host_shared.aph_info_config")

node_type = get_node_type()

APPLIANCE_INFO_FILE = "/etc/vmware/nsx/appliance-info.xml"
MPA_DECOM_STATE_FILE = "/etc/vmware/nsx-mpa/.decom_state"

_APH_IP_ADDR_KEY = "ip_address"
_APH_FQDN_KEY = "fqdn"
_APH_PORT_KEY = "port"
_APH_CERT_KEY = "certificate"
_APH_SERVICE_ENDPOINT_UUID_KEY = "service_endpoint_uuid"
_APH_ENTITIES_KEY = "entities_hosted"
_APH_ENTITY_TYPE_KEY = "entity_type"
_APH_ENTITY_UUID_KEY = "entity_uuid"
_APH_SSL_ENABLED_KEY = "ssl-enabled"

_APH_XML_IP_ADDR_KEY = "ip"
_APH_XML_FQDN_KEY = "fqdn"
_APH_XML_PORT_KEY = "port"
_APH_XML_CERT_KEY = "certificate"
_APH_XML_SERVICE_ENDPOINT_UUID_KEY = "uuid"
_APH_XML_ENTITIES_KEY = "entities"
_APH_XML_ENTITY_KEY = "entity"
_APH_XML_ENTITY_TYPE_KEY = "type"
_APH_XML_ENTITY_UUID_KEY = "uuid"
_APH_XML_SSL_ENABLED_KEY = "ssl-enabled"


def _update_legacy_mpa_decom_state():
    MPA_CONFIG_USER = "mpa"
    MPA_CONFIG_GROUP = "mpa"
    uid = gid = 0

    if node_type != NODE_TYPE_HOST_ESX and (not is_windows()):
        try:
            from pwd import getpwnam
            from grp import getgrnam
            uid = getpwnam(MPA_CONFIG_USER).pw_uid
            gid = getgrnam(MPA_CONFIG_GROUP).gr_gid
        except KeyError as ex:
            # This should never happen since the mpa user/group are created by
            # package install scripts.
            lg.warning("Could not find mpa user/group")
            uid = gid = 0

    try:
        with open(FilePaths.MPA_DECOM_STATE_FILE, "r") as fp:
            lg.info("Legacy MPA decom state file already exists. Skipping update.")
    except IOError as ex:
        if ex.errno == errno.ENOENT:
            with os.fdopen(os.open(FilePaths.MPA_DECOM_STATE_FILE,
                           os.O_WRONLY|os.O_CREAT, 0o644), 'w') as fp:
                if node_type != NODE_TYPE_HOST_ESX and (not is_windows()):
                    os.chown(FilePaths.MPA_DECOM_STATE_FILE, uid, gid)
            lg.info("Updated legacy MPA decom state")
        else:
            lg.warning("Error updating legacy mpa decom state")


def _get_aph_info_from_listener(node):
    if not ((_APH_IP_ADDR_KEY in node or _APH_FQDN_KEY in node) and
           _APH_PORT_KEY in node and _APH_SERVICE_ENDPOINT_UUID_KEY in node):
        return None

    _aph = {}
    if _APH_IP_ADDR_KEY in node:
        _aph[_APH_IP_ADDR_KEY] = node[_APH_IP_ADDR_KEY]
    if _APH_FQDN_KEY in node:
        _aph[_APH_FQDN_KEY] = node[_APH_FQDN_KEY]
    _aph[_APH_PORT_KEY] = node[_APH_PORT_KEY]

    if _APH_CERT_KEY in node and len(node[_APH_CERT_KEY]) > 0:
        _aph[_APH_CERT_KEY] = node[_APH_CERT_KEY]
        _aph[_APH_SSL_ENABLED_KEY] = "true"

    _aph[_APH_SERVICE_ENDPOINT_UUID_KEY] = node[_APH_SERVICE_ENDPOINT_UUID_KEY]
    _e_list = []
    if _APH_ENTITIES_KEY in node:
        for _entity in node[_APH_ENTITIES_KEY]:
            _e = {
                _APH_ENTITY_TYPE_KEY: _entity[_APH_ENTITY_TYPE_KEY],
                _APH_ENTITY_UUID_KEY: _entity[_APH_ENTITY_UUID_KEY]
            }
            _e_list.append(_e)
    _aph[_APH_ENTITIES_KEY] = _e_list
    return _aph


class AphInfo(object):
    def __init__(self, ip=None, port=None, uuid=None, fqdn=None, entities=[]):
        self.ip = ip
        self.port = port
        self.uuid = uuid
        self.fqdn = fqdn
        self.entities = entities

    @property
    def is_master(self):
        return any([e.is_master for e in self.entities])


class EntityInfo(object):
    def __init__(self, type=None, uuid=None, is_master=False):
        self.type = type
        self.uuid = uuid
        self.is_master = is_master


def _read_appliance_info_xml():
    """Parse appliance_info.xml and return aph info list
    """
    def ParseAphInfo(aph_info):
        # TODO Certificate parsing
        ip = _get_node_text(aph_info, 'ip')
        port = _get_node_text(aph_info, 'port')
        uuid = _get_node_text(aph_info, 'uuid')
        fqdn = _get_node_text(aph_info, 'fqdn')
        entities = []
        for entity in aph_info.findall('./entities/entity'):
            e_type = _get_node_text(entity, 'type')
            e_uuid = _get_node_text(entity, 'uuid')
            master = _get_node_text(entity, 'master')
            entities.append(EntityInfo(e_type, e_uuid, master == 'true'))
        return AphInfo(ip, port, uuid, fqdn, entities)

    def _get_node_text(elem, node_name):
        node = elem.find(node_name)
        return node.text if node is not None else None

    try:
        tree = ET.parse(FilePaths.APPLIANCE_INFO_FILE)
        list_root = tree.getroot().find('appliance-proxies')

        aph_list = []
        for aph_info in list_root.findall('appliance-proxy'):
            aph_list.append(ParseAphInfo(aph_info))
        return aph_list
    except ET.ParseError:
        # Appliance info file will be empty while transferring rmq connection to nsxrpc
        lg.warning("Invalid XML file: %s, ignoring", FilePaths.APPLIANCE_INFO_FILE)
        return []
    except Exception:
        lg.exception("Error reading " + FilePaths.APPLIANCE_INFO_FILE,
                     error_code=ec.INTERNAL_ERROR)
        return []


def _write_appliance_info_xml(aph_list=[]):
    """Persist APH info in appliance-info.xml
    """
    aph_json_key_to_xml_key = {
        _APH_IP_ADDR_KEY : _APH_XML_IP_ADDR_KEY,
        _APH_FQDN_KEY : _APH_XML_FQDN_KEY,
        _APH_PORT_KEY : _APH_XML_PORT_KEY,
        _APH_CERT_KEY : _APH_XML_CERT_KEY,
        _APH_SSL_ENABLED_KEY : _APH_XML_SSL_ENABLED_KEY,
        _APH_SERVICE_ENDPOINT_UUID_KEY : _APH_XML_SERVICE_ENDPOINT_UUID_KEY,
        _APH_ENTITIES_KEY : _APH_XML_ENTITIES_KEY,
        _APH_ENTITY_TYPE_KEY : _APH_XML_ENTITY_TYPE_KEY,
        _APH_ENTITY_UUID_KEY : _APH_XML_ENTITY_UUID_KEY,
    }

    def AddSubElement(root, node_name, dic):
        if node_name in aph_json_key_to_xml_key:
            node_handle = ET.SubElement(root, aph_json_key_to_xml_key[node_name])
        else:
            node_handle = ET.SubElement(root, node_name)

        if dic != None and node_name in dic:
            node_handle.text = str(dic[node_name])
        return node_handle

    # Create empty tree and add appliance-proxy entries from aph_list
    aph_info = ET.Element('appliance-info')
    aph_list_root = AddSubElement(aph_info, 'appliance-proxies', None)
    for aph in aph_list:
        aph_root = AddSubElement(aph_list_root, 'appliance-proxy', None)

        AddSubElement(aph_root, _APH_SERVICE_ENDPOINT_UUID_KEY, aph)
        if _APH_IP_ADDR_KEY in aph:
            AddSubElement(aph_root, _APH_IP_ADDR_KEY, aph)
        if _APH_FQDN_KEY in aph:
            AddSubElement(aph_root, _APH_FQDN_KEY, aph)
        AddSubElement(aph_root, _APH_PORT_KEY, aph)
        if _APH_CERT_KEY in aph:
            AddSubElement(aph_root, _APH_CERT_KEY, aph)
            AddSubElement(aph_root, _APH_SSL_ENABLED_KEY, aph)

        if _APH_ENTITIES_KEY in aph:
            entities_root = AddSubElement(aph_root, _APH_ENTITIES_KEY, None)
            for entity in aph[_APH_ENTITIES_KEY]:
                entity_root = AddSubElement(entities_root, _APH_XML_ENTITY_KEY,
                                            None)
                AddSubElement(entity_root, _APH_ENTITY_TYPE_KEY, entity)
                AddSubElement(entity_root, _APH_ENTITY_UUID_KEY, entity)

    try:
        xml_str = minidom.parseString(
                     ET.tostring(aph_info)).toprettyxml(indent=' ' * 4)

        # If aph_list is empty here, we are in the cleanup path. If the
        # appliance info file contains valid xml, then we write out empty tags
        # to the file if aph_list length is 0. If not, we do not update the
        # appliance info file. An invalid appliance info file is an indicator
        # of a state where we are connected to an older version MP.
        try:
           minidom.parse(FilePaths.APPLIANCE_INFO_FILE)
        except:
           if len(aph_list) == 0:
               return True

        with open(FilePaths.APPLIANCE_INFO_FILE, "w") as fp:
            fp.write(xml_str)

        # Create the decom state cache file. Failure is ok, as the old
        # MPA will check the validity of appliance info file as well.
        if len(aph_list) > 0:
            _update_legacy_mpa_decom_state()
    except Exception:
        lg.exception("Error updating " + FilePaths.APPLIANCE_INFO_FILE,
                     error_code=ec.INTERNAL_ERROR)
        return False
    return True
