# **********************************************************************
# Copyright 2015-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import nsx_logging
import os
try:
    import simplejson
except ImportError:
    import json as simplejson
import uuid
from cli.utils.command_utils import invoke_command
from cli.utils.error_code import ErrorCode as ec

lg = nsx_logging.getLogger("cli.utils.edge_utils")

_SUDO = "/usr/bin/sudo"
_EDGE_APPCTL_BIN = "/usr/local/bin/edge-appctl"
_OPSAGENT_APPCTL_BIN = "/usr/bin/nsx-appctl"
_EDGE_MDPROXY_BIN = "/opt/vmware/nsx-edge/bin/mdproxy.py"
_OPSAGENT_CTL = "/var/run/vmware/nsx-opsagent/opsagent-cli"
_SERVICE_DICT = {
    'dpd': {
        'name': 'dataplane',
        'socket': '/var/run/vmware/edge/dpd.ctl'
        },
    'nsxa': {
        'name': 'local-controller',
        'socket': '/var/run/vmware/edge/nsxa.ctl'
        },
    'dhcp_backend': {
        'name': 'dhcp',
        'socket': '/var/run/vmware/edge/dhcp.ctl'
        },
    'lldp': {
        'name': 'lldp',
        'socket': '/var/run/vmware/nsx-opsagent/opsagent-cli'
        },
    'lb': {
        'name': 'lb',
        'script':'/opt/vmware/nsx-edge/bin/lb-cli.py'
        },
    'lb_perf': {
        'name': 'lb_perf',
        'script':'/opt/vmware/nsx-edge/bin/lb-perf-config.py'
        },
    'iked': {
        'name': 'iked',
        'socket': '/var/run/vmware/edge/ike.ctl'
        },
    'dns-forwarder': {
        'name': 'dns-forwarder',
        'script':'/opt/vmware/nsx-edge/bin/dns/dns_fdr-cli.py'
        },
    'rcpm': {
        'name': 'rcpm',
        'socket': '/var/run/vmware/edge/rcpm.ctl'
        }
    }


def invoke_edge_script_command_raw(cmd_list, sendto, is_sudo=False):
    """Invoke an edge script command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
        sendto: lb = send to load balancer

    Return:
        (returncode, output, error) tuple.
    """
    cmd = [_SERVICE_DICT[sendto]['script']]
    cmd.extend(cmd_list)
    if is_sudo:
        cmd.insert(0, _SUDO)
    rc, out, err = invoke_command(cmd)
    return rc, out, err


def invoke_edge_appctl_command_raw(cmd_list, sendto='dpd'):
    """Invoke an edge-appctl command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["lswitch/show", "<uuid>"]).
        sendto: dpd = send to datapath
                nsxa = send to nsx-agent
                mpa = send to mpa
    Return:
        (returncode, output, error) tuple.
    """
    cmd = [_EDGE_APPCTL_BIN, "-t", _SERVICE_DICT[sendto]['socket']]
    cmd.extend(cmd_list)
    rc, out, err = invoke_command(cmd)
    return rc, out, err


def invoke_edge_appctl_command(cmd_list, sendto='dpd'):
    """Invoke an edge-appctl command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["lswitch/show", "<uuid>"]).
        sendto: dpd = send to datapath
                nsxa = send to nsx-agent
                mpa = send to mpa
    Return:
        (returncode, output, err) tuple.
            returncode will be 0 on success, else non-zero:
            On error, err will contain an error message suitable for
            display on the CLI.
    """
    rc, out, err = invoke_edge_appctl_command_raw(cmd_list, sendto)
    if rc != 0:
        lg.error("Failed to invoke edge-appctl command: %s", err,
                 error_code=ec.OPERATION_FAILED)
        substring_list = ["No such file or directory",
                          "cannot connect to"]
        if any(substring in err for substring in substring_list):
            return (rc, None,
                    ("The %s service is in error state, has failed or "
                     "is disabled" % _SERVICE_DICT[sendto]['name']))
        else:
            # Return the first line of error text from edge-appctl.
            errmsg = err.splitlines()[0] if err else "An internal error occurred"
            return (rc, None, errmsg)
    return (rc, out, None)


def invoke_edge_mdproxy_command_raw(cmd_list):
    """Invoke an edge mdproxy command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["--show-all", "all"]).
    Return:
        (returncode, output, error) tuple.
    """
    cmd = [_EDGE_MDPROXY_BIN]
    cmd.extend(cmd_list)
    rc, out, err = invoke_command(cmd)
    return rc, out, err


def invoke_edge_mdproxy_command(cmd_list):
    """Invoke an edge mdproxy command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["--show-all", "all"]).
    Return:
        (returncode, output, err) tuple.
            returncode will be 0 on success, else non-zero:
            On error, err will contain an error message suitable for
            display on the CLI.
    """
    rc, out, err = invoke_edge_mdproxy_command_raw(cmd_list)
    if rc != 0:
        lg.error("Failed to invoke edge mdproxy command: %s", err,
                 error_code=ec.OPERATION_FAILED)
        return (rc, None, err)
    return (rc, out, None)


def invoke_edge_lb_command_raw(cmd_list):
    """Invoke an edge lb command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["--get", "--lb-status"]).
    Return:
        (returncode, output, error) tuple.
    """
    rc, out, err = invoke_edge_script_command_raw(cmd_list, 'lb')
    return rc, out, err



def invoke_edge_lb_command(cmd_list):
    """Invoke an edge lb command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["--get", "--lb-status"]).
    Return:
        (returncode, output, err) tuple.
            returncode will be 0 on success, else non-zero:
            On error, err will contain an error message suitable for
            display on the CLI.
    """
    rc, out, err = invoke_edge_lb_command_raw(cmd_list)
    if rc != 0:
        lg.error("Failed to invoke edge lb command: %s", err)
        return (rc, None, err)
    return (rc, out, None)


def invoke_edge_lb_perf_command_raw(cmd_list, is_sudo=True):
    """Invoke an edge lb command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["apply", "--profile", "http"]).
    Return:
        (returncode, output, error) tuple.
    """
    rc, out, err = invoke_edge_script_command_raw(cmd_list, 'lb_perf', is_sudo)
    return rc, out, err


def invoke_edge_lb_perf_command(cmd_list, is_sudo=True):
    """Invoke an edge lb perf command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["apply", "--profile", "http"]).
    Return:
        (returncode, output, err) tuple.
            returncode will be 0 on success, else non-zero:
            On error, err will contain an error message suitable for
            display on the CLI.
    """
    rc, out, err = invoke_edge_lb_perf_command_raw(cmd_list, is_sudo)
    if rc != 0:
        lg.error("Failed to invoke edge lb perf command: %s", err)
        return (rc, None, err)
    return (rc, out, None)


def invoke_edge_dns_fdr_command_raw(cmd_list, is_sudo=True):
    """Invoke an edge dns_fdr command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["--show-status", "all"]).
    Return:
        (returncode, output, error) tuple.
    """
    rc, out, err = invoke_edge_script_command_raw(
        cmd_list, 'dns-forwarder', is_sudo)
    return rc, out, err



def invoke_edge_dns_fdr_command(cmd_list, is_sudo=True):
    """Invoke an edge dns_fdr command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["--show-status", "all"]).
    Return:
        (returncode, output, err) tuple.
            returncode will be 0 on success, else non-zero:
            On error, err will contain an error message suitable for
            display on the CLI.
    """
    rc, out, err = invoke_edge_dns_fdr_command_raw(cmd_list, is_sudo)
    if rc != 0:
        lg.error("Failed to invoke edge dns_fdr command: %s", err)
        return (rc, None, err)
    return (rc, out, None)


def invoke_opsagent_command(cmd_list):
    """Invoke opsAgent command.

    Args:
        cmd_list: List containing command to run and arguments to pass,
                  (e.g ["edgesvc/stats/get"]).
    Return:
        (returncode, output, err) tuple.
            returncode will be 0 on success, else non-zero:
            On error, err will contain an error message suitable for
            display on the CLI.
    """
    cmd = [_OPSAGENT_APPCTL_BIN, "-t", _OPSAGENT_CTL]
    cmd.extend(cmd_list)
    rc, out, err = invoke_command(cmd)
    return rc, out, err


def send_mpa_config_change_event(event_type):
    """Send config change event to MPA.

    Args:
        event_type: Either 'interface' or 'maintenance_mode'
    """
    cmd = ['edgesvc/cli/change', event_type]
    rc, _, err = invoke_opsagent_command(cmd)
    if rc != 0:
        lg.warn("Failed to send config change event to MPA: %s", err)


def get_edge_dp_physical_interfaces(sendto="dpd"):
    """Retrieve a list of physical interfaces from edge dp

    Return:
        List of interfaces information
    """
    cmd = ["physical_port/show"]
    rc, out, err = invoke_edge_appctl_command(cmd, sendto)
    if rc != 0:
        lg.error("Error getting physical interfaces from datapath. " \
                 "rc: %d, out: %s, err: %s", rc, out, err)
        return None
    interfaces = simplejson.loads(out)
    return interfaces


def get_edge_interfaces(sendto="dpd"):
    """Retrieve a list of physical port names or logical port uuids

    Return:
        List of interface names.
    """
    entities = []
    # Get physical port name
    cmd = ["device_name/match", "bond,pnic", ""]
    rc, out, err = invoke_edge_appctl_command(cmd, sendto)
    if rc == 0:
        entities += simplejson.loads(out)
    # Get logical port uuid
    cmd = ["uuid/match", "lrouter-port,lswitch-port,tunnel-port,ipsec-tunnel", ""]
    rc, out, err = invoke_edge_appctl_command(cmd, sendto)
    if rc == 0:
        entities += simplejson.loads(out)
    return entities


def get_linux_nics():
    result = []
    cmd = ["/sbin/ifconfig", "-s", "-a"]
    rc, out, err = invoke_command(cmd)

    if rc != 0:
        lg.warn("Failed to get interface names: %s", err)
        return result

    lines = out.split("\n")
    if lines:
        lines = lines[1:]
    for line in lines:
        parts = line.split()
        if not parts or len(parts) < 2:
            continue
        result.append(parts[0])
    return result


def convert_to_uuid(uuid_obj):
    int_value = (uuid_obj["left"] << 64) + uuid_obj["right"]
    return str(uuid.UUID(int=int_value))


def get_vrf_name_by_lr_id(lr_id):
    """Get the vrf_name by vrf_id"""

    cmd_list = ["vrf-config/show"]
    rc, out, err = invoke_edge_appctl_command(cmd_list, 'rcpm')
    if rc != 0:
        return None

    vrfs = simplejson.loads(out)
    for vrf in vrfs.get('vrf_configs', []):
        if str(lr_id) == str(vrf.get('vrf_id', "")):
            return vrf['vrf_name']
    return None


def get_vrf_name_by_sr_uuid(sr_uuid):
    """Get the vrf_name by sr_uuid"""

    cmd_list = ["lr-config/show"]
    rc, out, err = invoke_edge_appctl_command(cmd_list, 'rcpm')
    if rc != 0:
        return None

    vrfs = simplejson.loads(out)
    for vrf in vrfs.get('lr_config', []):
        if str(sr_uuid) == str(vrf.get('id', "")):
            return vrf['vrf_name']
    return None


def get_locale_mode_for_sr(sr_uuid):
    """Get the locale_mode i.e. site-role for SR"""

    if sr_uuid is None:
        return None

    cmd_list = ["service-router/show", sr_uuid]
    rc, out, err = invoke_edge_appctl_command(cmd_list, sendto="nsxa")
    if rc != 0:
        lg.error("Error getting service router information from edge-agent." \
                 " rc: %d, out: %s, err: %s", rc, out, err)
        return None

    sr = simplejson.loads(out)
    if "lr_uuid" in sr:
        cmd_list = ["sr-cluster/show", sr["lr_uuid"]]
        rc, out, err = invoke_edge_appctl_command(cmd_list, sendto="nsxa")
        if rc != 0:
            lg.error("Error getting SR cluster information from edge-agent. "\
                     "rc: %d, out: %s, err: %s", rc, out, err)
            return None

        src = simplejson.loads(out)

        if "locale_ops_mode" in src:
            return src["locale_ops_mode"]

    return None
