# **********************************************************************
# Copyright 2014 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
try:
    import simplejson
except ImportError:
    import json as simplejson
import socket
import time

from cli.utils.exceptions import ApiException
from cli.utils.rest_utils import rest_request
from compat23 import compat23_httplib


NAPI_HOST = "127.0.0.1:7441"
RETRY_INTERVAL = 1
RETRY_MAX_TIME = 15


def get(path, headers=None):
    """Perform an HTTP GET operation
       Returns a tuple <response>, <body>, where <response>
       is the HTTP response code and <body> is the response body."""
    conn = None
    try:
        conn = compat23_httplib.HTTPConnection(NAPI_HOST)
        conn.request("GET", path, None, headers)
        resp = conn.getresponse()
        return resp.status, resp.read()
    finally:
        if conn:
            conn.close()


def _get_json_body(resp):
    ctype = resp.getheader("content-type", "")
    if ctype.lower() == "application/json":
        return simplejson.loads(resp.read())
    return None


def _http_op(op, path, body, headers, retry_time=RETRY_MAX_TIME):
    conn = None
    if not headers:
        headers = {}
    if body:
        headers["Content-Type"] = "application/json"
    start = time.time()
    while True:
        try:
            conn = compat23_httplib.HTTPConnection(NAPI_HOST)
            json_body = simplejson.dumps(body) if body else None
            conn.request(op, path, json_body, headers)
            resp = conn.getresponse()
            return resp.status, _get_json_body(resp)
        except socket.error as ex:
            if time.time() - start > retry_time:
                raise
            time.sleep(RETRY_INTERVAL)
        finally:
            if conn:
                conn.close()


def get_json(path, headers=None, retry_time=RETRY_MAX_TIME):
    """Perform an HTTP GET operation, where that operation is
       expected to return a json body. Returns a tuple
       <response>, <obj>, where <response> is the HTTP response
       code and <obj> is the converted json body (which may
       be None if the call fails)."""
    return _http_op("GET", path, None, headers, retry_time=retry_time)


def put_json(path, body, headers=None, retry_time=RETRY_MAX_TIME):
    """Perform an HTTP PUT operation, using the
       given body object. The body is converted to json
       before sending to the server. The call is expected to
       return a json body. Returns a tuple <response>, <obj>,
       where <response> is the HTTP response code and <obj>
       is the converted json body (which may be None if the
       call fails)."""
    return _http_op("PUT", path, body, headers, retry_time=retry_time)


def post_json(path, body, headers=None, retry_time=RETRY_MAX_TIME):
    """Perform an HTTP POST operation, posting the
       given body object. The body is converted to json
       before posting to the server. The call is expected to
       return a json body. Returns a tuple <response>, <obj>,
       where <response> is the HTTP response code and <obj>
       is the converted json body (which may be None if the
       call fails)."""
    return _http_op("POST", path, body, headers, retry_time=retry_time)


def request(http_method, path, body=None, headers=None,
            retry_time=RETRY_MAX_TIME, retry_interval=RETRY_INTERVAL,
            retry_if_service_unavailable=False):
    return rest_request(NAPI_HOST, http_method, path, body, headers, retry_time,
                        retry_interval, retry_if_service_unavailable)
