# Copyright (C) 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import os
import json
import traceback
import sha.config
from nsx_utils.node_utils import get_node_type
from sha.contrib.metric.abc_base import Metric
from sha.contrib.metric.utils import run_command
from sha.contrib.metric.utils import timeout_cache

LB_CMD_EXECUTE_MAX_SEC = 60

class LbCapacity:
    factor_capacity = None

    @classmethod
    def get(cls):
        if cls.factor_capacity == None:
            if os.path.exists('/usr/sbin/virt-what'):
                cpu_count2factor = {2 : {'SMALL' : 1,
                                        'MEDIUM' : 0,
                                        'LARGE' : 0,
                                        'XLARGE' : 0},
                                    4 : {'SMALL' : 10,
                                         'MEDIUM' : 1,
                                         'LARGE' : 0,
                                         'XLARGE' : 0},
                                    8 : {'SMALL' : 40,
                                        'MEDIUM' : 4,
                                        'LARGE' : 1,
                                        'XLARGE' : 0},
                                    16 :{'SMALL' : 80,
                                        'MEDIUM' : 8,
                                        'LARGE' : 2,
                                        'XLARGE' : 1}}
                if os.cpu_count() in cpu_count2factor.keys():
                    cls.factor_capacity = cpu_count2factor[os.cpu_count()]
                else:
                    logger.info("Failed to get load balancer status {}".format(e))
                    cls.factor_capacity = cpu_count2factor[2]
            else:
                cls.factor_capacity = {'SMALL' : 750,
                                        'MEDIUM' : 75,
                                        'LARGE' : 18,
                                        'XLARGE' : 9}

        return cls.factor_capacity


class LbStatus(Metric):
    def __init__(self):
        self._logger = sha.config.get_logger(__name__)

    def get_lb_sr_id(self, lb_id):
        try:
            cmd = ['nsxcli', '-c', 'get', 'load-balancer', lb_id, '|', 'json']
            output = run_command(cmd, shell=False, timeout=LB_CMD_EXECUTE_MAX_SEC, env={'PYTHONPATH':''})
            lb_info = json.loads(output)
            return lb_info[0]['applied_to']['service_router_id']
        except Execption as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get load-balancer {}, {}".format(e, trace))
            return None

    def get_sr_info(self, sr_id):
        try:
           cmd = ['nsxcli', '-c', 'get', 'logical-router', sr_id, '|', 'json']
           output = run_command(cmd, shell=False, timeout=LB_CMD_EXECUTE_MAX_SEC, env={'PYTHONPATH':''})
           sr_info = json.loads(output)
           if 'SERVICE_ROUTER_TIER0' in sr_info:
               return {'sr_name' : sr_info['SERVICE_ROUTER_TIER0']['name'],
                       'sr_type' : 'T0'}
           else:
               return {'sr_name' : sr_info['SERVICE_ROUTER_TIER1']['name'],
                       'sr_type' : 'T1'}
        except Execption as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get logical-router {}, {}".format(e, trace))
            return None

    def get_sr_name(self, lb_status_list):
        for lb_status in lb_status_list['lbs']:
            lb_id = lb_status['uuid']
            sr_id = self.get_lb_sr_id(lb_id)
            sr_info = self.get_sr_info(sr_id)
            lb_status['sr_info'] = sr_info

    def get_lb_status(self):
        try:
            cmd = ['/opt/vmware/nsx-edge/bin/lb-cli.py', '--get', '--lb-status', '-d']
            output = run_command(cmd, shell=False, timeout=LB_CMD_EXECUTE_MAX_SEC, env={'PYTHONPATH':''})
            lb_status_list = json.loads(output)
            self.get_sr_name(lb_status_list)
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get load balancer status {}, {}".format(e, trace))
            return None

        return None if not lb_status_list else {'lb_type' : 'lb', 'context' : lb_status_list,
                        'lb_capacity' : LbCapacity.get(), 'source' : self.wavefront_source}

    @property
    def wavefront_source(self):
        # Get hostname as source to export metrics to wavefront
        return sha.config.HOST_NAME.lower()

    def get_dlb_status(self):
        try:
            cmd = ['/bin/vsipioctl', 'getlbstatus', '-n']
            output = run_command(cmd, shell=False, timeout=LB_CMD_EXECUTE_MAX_SEC)
            lb_status_list = json.loads(output)
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get dlb status {}, {}".format(e, trace))
            return None

        return None if not lb_status_list else {'lb_type' : 'dlb', 'context' : lb_status_list}

    @timeout_cache(10)
    def usage(self, *args, **kwargs):
        _NODE_TYPE = get_node_type();
        if _NODE_TYPE == sha.config.DIST_EDGE:
            return self.get_lb_status()
        elif _NODE_TYPE == sha.config.DIST_ESX:
            return self.get_dlb_status()
        else:
            return None
